(define-module (alterator ensign)
  :use-module (srfi srfi-1)
  :use-module (srfi srfi-2)
  :use-module (srfi srfi-13)
  :use-module (alterator object)
  :use-module (alterator str)
  :use-module (alterator plist)
  :use-module (alterator algo)
  :use-module (alterator woo)

  :use-module (alterator exit-handler)

  ;;backend variants
  :use-module (alterator ensign backend2)
  :use-module (alterator ensign backend3)
  :use-module (alterator ensign ensign)

  :export (ensign
           ensign-gate))

(define *mariners* (make-cell '()))

;; cleanup backend processes at finish

(define (stop-mariners . ignore)
  (define-operation quit #t)
  (cell-set! *mariners* (filter
                         (lambda(x) (not (quit (cadr x))))
                         (cell-ref *mariners*))))

(sigaction SIGHUP stop-mariners) ;;at sighup reset
(at-exit stop-mariners); at exit reset

;;low level backend processing

(define *clean-interval* (make-cell (* 60 60 5)))

(define (sweep-mariner)
  (define-operation quit #t)
  (let ((clean-interval (cell-ref *clean-interval*))
        (now (current-time)))
    (cell-set! *mariners*
               (filter
                (lambda(x)
                  (or (< (- now (cddr x)) clean-interval)
                      (not (quit (cadr x)))))
                (cell-ref *mariners*)))))

(define (load-mariner sym)
  (or (make-backend2 sym)
      (make-backend3 sym)))

(define (remember-mariner sym backend)
  (and backend
       (cell-set! *mariners*
                  (acons sym (cons backend (current-time)) (cell-ref *mariners*))))
  backend)

(define (forget-mariner sym backend)
  (define-operation quit #t)
  (quit backend)
  (cell-set! *mariners* (alist-delete sym (cell-ref *mariners*))))

(define (touch-mariner x)
    (set-cdr! (cdr x) (current-time))
    (sweep-mariner)
    (cadr x))

(define (resolv-mariner sym)
  (cond
   ((assoc sym (cell-ref *mariners*)) => touch-mariner)
   (else (remember-mariner sym (load-mariner sym)))))

(define (mariner-error . args)
 (list 'error (string-concatenate args)))

(define (run-mariner mariner action objects options)
  (define (pair-return z) (and (list? z) z))
  (run-operation mariner
                 action
                 (lambda(mariner . args)
                   (or (pair-return (apply mariner mariner args))
		       (mariner-error "unknown action for backend: " (car objects))))
                 (cdr objects)
                 options))

(define (call-mariner cmd)
  (let* ((name (car cmd))
	 (args (cdr cmd))
	 (sym (car name))
	 (mariner (resolv-mariner sym)))
    (cond
      ((not mariner) (mariner-error "backend not found: " sym))
      ((plistq 'action args) =>
			     (lambda(action)
			       (catch 'mariner-error
				      (lambda()
					(run-mariner mariner (sure-symbol (cdr action))
						     (car cmd)
						     (cdr cmd)))
				      (lambda (key . args)
					(forget-mariner sym mariner)
					(mariner-error "backend-communication error: " (format #f "~S" args))))))
      (else (mariner-error "action field not found: " (format #f "~S" cmd))))))

;; highlevel backend processing

(define (error-answer? cmd) (and (pair? cmd) (eq? (car cmd) 'error)))
(define (quote-answer? cmd) (and (pair? cmd) (eq? (car cmd) 'quote)))
(define (list-answer? cmd)  (and (pair? cmd) (pair? (car cmd))))

(define (out-normal name options)
  (cons name options))
(define (out-error name reason)
  (out-normal (string-append "/error" name) `(reason ,reason)))
(define (out-list name options)
   (if (string? (car options))
       (out-normal (string-append name "/" (car options))
                   (cons* 'name (car options) (cdr options)))
       (out-normal name options)))

(define (prepare-name name)
  (if (pair? name)
      name
      (and (not-empty-string? name)
           (not (string-index name #\*))
           (string-cut-repeated name #\/))))

;;filter: split command name, convert args to associative list
(define (prepare-cmds cmds)
  (filter-map (lambda(cmd)
                (and-let* (((pair? cmd))
                           (name (prepare-name (car cmd)))
                           ((not (null? name)))
                           ((not (member ".." name))))
                          (cons name (cdr cmd))))
              cmds))

;; new-version: process commands from list and autoconvert results
(define (process-cmd cmd)
  (let ((name (string-append "/" (string-join (car cmd) "/")))
        (answer (call-mariner cmd)))
    (cond
     ((null? answer) '())
     ;;answer will be destroyed from output queue, use (quote) for really empty answers
     ((error-answer? answer) (list (out-error name (cadr answer))))
     ((quote-answer? answer) (list (cdr answer)))
     ((list-answer? answer)
      (map (lambda (x)
             (if (quote-answer? x)
                 (cdr x)
                 (out-list name x)))
             answer))
     ((pair? answer) (list (out-normal name answer)))
     (else
      (list (out-error name "unsupported answer type"))))))

(define (ensign-gate cmds)
  (with-fluids ((woo-gate ensign-gate))
               (append-map process-cmd
                           (prepare-cmds cmds))))

;;;main entry point

(define (ensign)
  ;;load default backends
  (let ((backend (make-ensign *mariners* *clean-interval*)))
    (cell-set! *mariners*
               (acons "ensign" (cons backend (current-time)) '() )))
  (lambda(cmds next)
    (next (ensign-gate cmds))))
